class ILError(Exception):
    """ IngeniaLink generic error. """
    pass


class ILCreationError(ILError):
    """
    IngeniaLink creation error.
    """
    pass


class ILValueError(ILError):
    """
    IngeniaLink value error.
    """
    pass


class ILTimeoutError(ILError):
    """
    IngeniaLink timeout error.
    """
    pass


class ILAlreadyInitializedError(ILError):
    """
    InteniaLink already initialized error.
    """
    pass


class ILMemoryError(ILError):
    """
    IngeniaLink memory error.
    """
    pass


class ILDisconnectionError(ILError):
    """
    IngeniaLink disconnection error.
    """
    pass


class ILAccessError(ILError):
    """
    IngeniaLink access error.
    """
    pass


class ILStateError(ILError):
    """
    IngeniaLink state error.
    """
    pass


class ILIOError(ILError):
    """
    IngeniaLink I/O error.
    """
    pass


class ILNotSupportedError(ILError):
    """
    IngeniaLink Not supported error.
    """
    pass


class ILWrongRegisterError(ILError):
    """
    IngeniaLink Wrong register error.
    """
    pass


class ILWrongCRCError(ILError):
    """
    IngeniaLink Wrong CRC error.
    """
    pass


class ILNACKError(ILError):
    """
    IngeniaLink NACK error.
    """

    
class UDPException(Exception):
    """
    Exception on UDP action.
    """
    pass



from .net import (Network, NetworkMonitor, devices, NET_PROT, NET_STATE,
                  NET_DEV_EVT)
from .servo import (Servo, lucky, SERVO_STATE, SERVO_FLAGS, SERVO_MODE,
                    SERVO_UNITS_TORQUE, SERVO_UNITS_POS, SERVO_UNITS_VEL,
                    SERVO_UNITS_ACC)
from .monitor import Monitor, MONITOR_TRIGGER
from .poller import Poller
from .registers import Register, REG_DTYPE, REG_ACCESS, REG_PHY
from .dict_ import Dictionary
from .dict_labels import LabelsDictionary
from .canopen.servo_node import Servo as CANOpenServo
from .canopen.net import Network as CANOpenNetwork
from .canopen.poller_node import Poller as CANOpenPoller
from .err import err_ipb_last
from ._ingenialink import lib
from ._utils import pstr

__all__ = ['Network', 'NetworkMonitor', 'devices', 'NET_PROT', 'NET_DEV_EVT',
           'NET_STATE',
           'Servo', 'lucky', 'SERVO_STATE', 'SERVO_FLAGS', 'SERVO_MODE',
           'SERVO_UNITS_TORQUE', 'SERVO_UNITS_POS', 'SERVO_UNITS_VEL',
           'SERVO_UNITS_ACC',
           'Monitor', 'MONITOR_TRIGGER',
           'Poller',
           'Register', 'REG_DTYPE', 'REG_ACCESS', 'REG_PHY',
           'Dictionary',
           'LabelsDictionary',
           'CANOpenNetwork', 'CAN_DEVICE', 'CANOpenPoller', 'CANOpenServo',
           'err_ipb_last']

__version__ = '5.2.2'

try:
    __ingenialink_C_version__ = pstr(lib.il_version())
except Exception as e:
    __ingenialink_C_version__ = '-'



import collections

from ._ingenialink import ffi, lib
from ._utils import cstr, pstr, raise_null, raise_err

from .registers import Register, REG_DTYPE
from .dict_labels import LabelsDictionary

import xml.etree.ElementTree as ET


class SubCategories(object):
    """Sub-categories.

    Args:
        dict_ (il_dict_t *): Ingenia dictionary instance.
        cat_id (str): Category ID (parent).
    """

    def __init__(self, dict_, cat_id):
        self._dict = dict_
        self._cat_id = cat_id

        self._load_scat_ids()

    def _load_scat_ids(self):
        """
        Load sub-category IDs from dictionary.
        """
        scat_ids = lib.il_dict_scat_ids_get(self._dict, cstr(self._cat_id))

        self._scat_ids = []
        i = 0
        scat_id = scat_ids[0]
        while scat_id != ffi.NULL:
            self._scat_ids.append(pstr(scat_id))
            i += 1
            scat_id = scat_ids[i]

        lib.il_dict_scat_ids_destroy(scat_ids)

    @property
    def scat_ids(self):
        """
        Obtain all sub-category identifiers.

        Returns:
            list: Sub-category identifiers.
        """
        return self._scat_ids

    def labels(self, scat_id):
        """
        Obtain labels for a certain sub-category identifiers.

        Returns:
            dict: Labels dictionary.
        """
        labels_p = ffi.new('il_dict_labels_t **')
        r = lib.il_dict_scat_get(self._dict, cstr(self._cat_id), cstr(scat_id),
                                 labels_p)
        raise_err(r)

        return LabelsDictionary._from_labels(labels_p[0])


class Categories(object):
    """
    Categories.

    Args:
        dict_ (il_dict_t *): Ingenia dictionary instance.
    """

    def __init__(self, dict_):
        self._dict = dict_

        self._load_cat_ids()

    def _load_cat_ids(self):
        """
        Load category IDs from dictionary.
        """
        cat_ids = lib.il_dict_cat_ids_get(self._dict)

        self._cat_ids = []
        i = 0
        cat_id = cat_ids[0]
        while cat_id != ffi.NULL:
            self._cat_ids.append(pstr(cat_id))
            i += 1
            cat_id = cat_ids[i]

        lib.il_dict_cat_ids_destroy(cat_ids)

    @property
    def cat_ids(self):
        """
        Obtain all Category Identifiers.

        Returns:
            list: Category IDs."""
        return self._cat_ids

    def labels(self, cat_id):
        """
        Obtain labels for a certain category ID.

        Returns:
            dict: Labels dictionary.
        """
        labels_p = ffi.new('il_dict_labels_t **')
        r = lib.il_dict_cat_get(self._dict, cstr(cat_id), labels_p)
        raise_err(r)

        return LabelsDictionary._from_labels(labels_p[0])

    def scats(self, cat_id):
        """
        Obtain all sub-categories.

        Returns:
            SubCategories: Sub-categories.
        """
        return SubCategories(self._dict, cat_id)


class RegistersDictionary(collections.Mapping):
    """
    Registers dictionary.

    Args:
        dict_ (il_dict_t *): Ingenia dictionary instance.
    """

    def __init__(self, dict_, subnode):
        self._dict = dict_
        self._subnode = subnode

        self._load_reg_ids()

    def _load_reg_ids(self):
        """
        Load register IDs from dictionary.
        """
        self._ids = []
        ids = lib.il_dict_reg_ids_get(self._dict, self._subnode)

        i = 0
        _id = ids[0]
        while _id != ffi.NULL:
            self._ids.append(pstr(_id))
            i += 1
            _id = ids[i]

        lib.il_dict_reg_ids_destroy(ids)

    def __getitem__(self, _id):
        reg_p = ffi.new('il_reg_t **')
        r = lib.il_dict_reg_get(self._dict, cstr(_id), reg_p, self._subnode)
        raise_err(r)

        return Register._from_register(reg_p[0])

    def __len__(self):
        return len(self._ids)

    def __iter__(self):
        return iter(self._ids)


class Dictionary(object):
    """
    Ingenia Dictionary.

    Args:
        dict_f (str): Dictionary file name.

    Raises:
        ILCreationError: If the dictionary could not be created.
    """
    def __init__(self, dict_f):
        dict_ = lib.il_dict_create(cstr(dict_f))
        raise_null(dict_)

        # Dictionary version
        self._dict = ffi.gc(dict_, lib.il_dict_destroy)

        self._version = lib.il_dict_version_get(dict_)
        self._subnodes = lib.il_dict_subnodes_get(dict_)

        self._rdicts = []
        for subnode in range(0, self._subnodes):
            rdict = RegistersDictionary(self._dict, subnode)
            self._rdicts.append(rdict)
        self._cats = Categories(self._dict)

    @classmethod
    def _from_dict(cls, dict_):
        """
        Create a new class instance from an existing dictionary.
        """
        inst = cls.__new__(cls)
        inst._dict = dict_

        inst._version = lib.il_dict_version_get(inst._dict)
        inst._subnodes = lib.il_dict_subnodes_get(inst._dict)

        inst._rdicts = []
        for subnode in range(0, inst._subnodes):
            rdict = RegistersDictionary(inst._dict, subnode)
            inst._rdicts.append(rdict)
        inst._cats = Categories(inst._dict)

        return inst

    def version_get(self, dict_):
        """

        Args:
            dict_: Dictionary path.

        Returns:
            str: Dictionray version.
        """
        return lib.il_dict_version_get(dict_)

    def save(self, fname):
        """
        Save dictionary.

        Args:
            fname (str): Output file name/path.
        """
        r = lib.il_dict_save(self._dict, cstr(fname))
        raise_err(r)

    def get_regs(self, subnode):
        """
        Obtain all the registers of a subnode.

        Args:
            subnode: Subnode.

        Returns:
            array: List of registers.
        """
        if subnode < self._subnodes:
            return self._rdicts[subnode]

    # @property
    # def regs(self):
    #     """RegistersDictionary: Registers dictionary."""
    #     return self._rdict

    def reg_storage_update(self, id_, value):
        """
        Update register storage.

        Args:
            id_ (str): Register ID.
            value: Value.
        """
        reg = self.regs[id_]
        value_ = ffi.new('il_reg_value_t')

        if reg.dtype == REG_DTYPE.S8:
            value_.storage.s8 = int(value)
        elif reg.dtype == REG_DTYPE.U8:
            value_.storage.u8 = int(value)
        if reg.dtype == REG_DTYPE.S16:
            value_.storage.s16 = int(value)
        elif reg.dtype == REG_DTYPE.U16:
            value_.storage.u16 = int(value)
        if reg.dtype == REG_DTYPE.S32:
            value_.storage.s32 = int(value)
        elif reg.dtype == REG_DTYPE.U32:
            value_.storage.u32 = int(value)
        if reg.dtype == REG_DTYPE.S64:
            value_.storage.s64 = int(value)
        elif reg.dtype == REG_DTYPE.U64:
            value_.storage.u64 = int(value)
        elif reg.dtype == REG_DTYPE.FLOAT:
            value_.storage.flt = float(value)
        else:
            raise ValueError('Unsupported register data type')

        r = lib.il_dict_reg_storage_update(cstr(id_), value_)
        raise_err(r)

    @property
    def cats(self):
        """
        Obtains all categories of the dictionary.

        Returns:
            Categories: Categories.
        """
        return self._cats

    @property
    def version(self):
        """
        Obtain dictionary version.

        Returns:
            str: Version.
        """
        return pstr(self._version)

    @property
    def subnodes(self):
        """
        Obtain number of subnodes defined in the dictionary.

        Returns:
            int: Subnodes.
        """
        return self._subnodes



from enum import Enum

from ._ingenialink import lib, ffi
from . import exceptions as exc


def cstr(v):
    """
    Convert Python 3.x string to C compatible char *.
    """
    return v.encode('utf8')


def pstr(v):
    """
    Convert C string to Python 3.x compatible str.
    """
    convert = ""
    try:
        convert = ffi.string(v).decode('utf8')
    except Exception as e:
        print("Error converting C string to Python. Exception: {}".format(e))
    return convert


def to_ms(s):
    """
    Convert from seconds to milliseconds.

    Args:
        s (float, int): Value in seconds.

    Returns:
        int: Value in milliseconds.
    """
    return int(s * 1e3)


class INT_SIZES(Enum):
    """
    Integer sizes.
    """
    S8_MIN = -128
    S16_MIN = -32767 - 1
    S32_MIN = -2147483647 - 1
    S64_MIN = 9223372036854775807 - 1

    S8_MAX = 127
    S16_MAX = 32767
    S32_MAX = 2147483647
    S64_MAX = 9223372036854775807

    U8_MAX = 255
    U16_MAX = 65535
    U32_MAX = 4294967295
    U64_MAX = 18446744073709551615


def raise_null(obj):
    """
    Raise exception if object is ffi.NULL.

    Raises:
        ILCreationError: If the object is NULL.
    """
    if obj == ffi.NULL:
        msg = pstr(lib.ilerr_last())
        raise exc.ILCreationError(msg)


def raise_err(code, msg=None):
    """
    Raise exception if the code is non-zero.

    Raises:
        ILValueError: if code is lib.IL_EINVAL
        ILTimeoutError: if code is lib.IL_ETIMEDOUT
        ILAlreadyInitializedError: if code is lib.IL_EALREADY
        ILMemoryError: if code is lib.IL_ENOMEM
        ILDisconnectionError: if code is lib.IL_EDISCONN
        ILAccessError: if code is lib.IL_EACCESS
        ILStateError: if code is lib.IL_ESTATE
        ILError: if code is lib.IL_EFAULT
    """
    if code == 0:
        return

    # obtain message and raise its matching exception
    if not msg:
        msg = pstr(lib.ilerr_last())

    if code == lib.IL_EINVAL:
        raise exc.ILValueError(msg)
    elif code == lib.IL_ETIMEDOUT:
        raise exc.ILTimeoutError(msg)
    elif code == lib.IL_EALREADY:
        raise exc.ILAlreadyInitializedError(msg)
    elif code == lib.IL_ENOMEM:
        raise exc.ILMemoryError(msg)
    elif code == lib.IL_EDISCONN:
        raise exc.ILDisconnectionError(msg)
    elif code == lib.IL_EACCESS:
        raise exc.ILAccessError(msg)
    elif code == lib.IL_ESTATE:
        raise exc.ILStateError(msg)
    elif code == lib.IL_EIO:
        raise exc.ILIOError(msg)
    elif code == lib.IL_ENOTSUP:
        raise exc.ILNotSupportedError(msg)
    elif code == lib.IL_EWRONGREG:
        raise exc.ILWrongRegisterError(msg)
    elif code == lib.IL_EWRONGCRC:
        raise exc.ILWrongCRCError(msg)
    elif code == lib.IL_ENACK:
        raise exc.ILNACKError(msg)
    else:
        raise exc.ILError(msg)



from enum import Enum

from .._ingenialink import lib

from .._utils import *
from ..registers import REG_DTYPE, REG_ACCESS, REG_PHY


class Register(object):
    """ Register.

        Args:
            identifier (str): Identifier.
            units (str): Units.
            address (int): Address.
            dtype (REG_DTYPE): Data type.
            access (REG_ACCESS): Access type.
            phy (REG_PHY, optional): Physical units.
            subnode (int): Subnode
            storage (any, optional): Storage.
            range (tuple, optional): Range (min, max).
            labels (dict, optional): Register labels.
            cat_id (str, optional): Category ID.
            scat_id (str, optional): Sub-category ID.
            internal_use (int, optional): Internal use.

        Raises:
            TypeError: If any of the parameters has invalid type.
    """

    def __init__(self, identifier, units, cyclic, idx, subidx, dtype,
                 access, phy=REG_PHY.NONE, subnode=1, storage=None,
                 range=(None, None), labels={}, enums=[], enums_count=0,
                 cat_id=None, scat_id=None, internal_use=0):
        if not isinstance(dtype, REG_DTYPE):
            raise_err(lib.IL_EINVAL, 'Invalid data type')

        if not isinstance(access, REG_ACCESS):
            raise_err(lib.IL_EACCESS, 'Invalid access type')

        if not isinstance(phy, REG_PHY):
            raise_err(lib.IL_EINVAL, 'Invalid physical units type')

        # initialize register
        self.__identifier = identifier
        self.__units = units
        self.__idx = idx
        self.__subidx = subidx
        self.__subnode = subnode
        self.__cyclic = cyclic
        self.__dtype = dtype.value
        self.__access = access.value
        self.__phy = phy.value
        self.__internal_use = internal_use
        self.__storage = storage
        self.__storage_valid = 0 if not storage else 1
        self.__range = (None, None) if not range else range

        if dtype == REG_DTYPE.S8:
            if storage:
                self.__storage = int(storage)

            range_min = (range[0] if range[0] else INT_SIZES.S8_MIN.value)
            range_max = (range[1] if range[1] else INT_SIZES.S8_MAX.value)
            self.__range = (int(range_min), int(range_max))
        elif dtype == REG_DTYPE.U8:
            if storage:
                self.__storage = int(storage)

            range_min = range[0] if range[0] else 0
            range_max = (range[1] if range[1] else INT_SIZES.U8_MAX.value)
            self.__range = (int(range_min), int(range_max))
        if dtype == REG_DTYPE.S16:
            if storage:
                self.__storage = int(storage)

            range_min = (range[0] if range[0] else INT_SIZES.S16_MIN.value)
            range_max = (range[1] if range[1] else INT_SIZES.S16_MAX.value)
            self.__range = (int(range_min), int(range_max))
        elif dtype == REG_DTYPE.U16:
            if storage:
                self.__storage = int(storage)

            range_min = range[0] if range[0] else 0
            range_max = (range[1] if range[1] else INT_SIZES.U16_MAX.value)
            self.__range = (int(range_min), int(range_max))
        if dtype == REG_DTYPE.S32:
            if storage:
                self.__storage = int(storage)

            range_min = (range[0] if range[0] else INT_SIZES.S32_MIN.value)
            range_max = (range[1] if range[1] else INT_SIZES.S32_MAX.value)
            self.__range = (int(range_min), int(range_max))
        elif dtype == REG_DTYPE.U32:
            if storage:
                self.__storage = int(storage)

            range_min = range[0] if range[0] else 0
            range_max = (range[1] if range[1] else INT_SIZES.U32_MAX.value)
            self.__range = (int(range_min), int(range_max))
        if dtype == REG_DTYPE.S64:
            if storage:
                self.__storage = int(storage)

            range_min = (range[0] if range[0] else INT_SIZES.S64_MIN.value)
            range_max = (range[1] if range[1] else INT_SIZES.S64_MAX.value)
            self.__range = (int(range_min), int(range_max))
        elif dtype == REG_DTYPE.U64:
            if storage:
                self.__storage = int(storage)

            range_min = range[0] if range[0] else 0
            range_max = (range[1] if range[1] else INT_SIZES.U64_MAX.value)
            self.__range = (int(range_min), int(range_max))
        elif dtype == REG_DTYPE.FLOAT:
            if storage:
                self.__storage = float(storage)

            range_min = (range[0] if range[0] else INT_SIZES.S32_MIN.value)
            range_max = (range[1] if range[1] else INT_SIZES.S32_MAX.value)
            self.__range = (float(range_min), float(range_max))
        else:
            self.__storage_valid = 0

        self.__labels = labels
        self.__enums = []
        self.__enums_count = enums_count

        for enum in enums:
            for key, value in enum.items():
                dictionary = {
                    'label': value,
                    'value': int(key)
                }
                self.__enums.append(dictionary)

        self.__cat_id = cat_id
        self.__scat_id = scat_id

    @property
    def identifier(self):
        """ str: Register identifier """
        return self.__identifier

    @property
    def units(self):
        """ str: Register units """
        return self.__units

    @property
    def idx(self):
        """ int: Register index. """
        return self.__idx

    @property
    def subidx(self):
        """ int: Register subindex. """
        return self.__subidx

    @property
    def subnode(self):
        """ int: Register subnode. """
        return self.__subnode

    @property
    def cyclic(self):
        """ str: Register cyclic type. """
        return self.__cyclic

    @property
    def dtype(self):
        """ int: Register data type. """
        return REG_DTYPE(self.__dtype)

    @property
    def access(self):
        """ int: Register access type. """
        return REG_ACCESS(self.__access)

    @property
    def phy(self):
        """ int: Register physical units. """
        return REG_PHY(self.__phy)

    @property
    def storage(self):
        """ Register storage. """
        if not self.__storage_valid:
            return None

        if self.dtype == REG_DTYPE.S8:
            return self.__storage
        elif self.dtype == REG_DTYPE.U8:
            return self.__storage
        if self.dtype == REG_DTYPE.S16:
            return self.__storage
        elif self.dtype == REG_DTYPE.U16:
            return self.__storage
        if self.dtype == REG_DTYPE.S32:
            return self.__storage
        elif self.dtype == REG_DTYPE.U32:
            return self.__storage
        if self.dtype == REG_DTYPE.S64:
            return self.__storage
        elif self.dtype == REG_DTYPE.U64:
            return self.__storage
        elif self.dtype == REG_DTYPE.FLOAT:
            return self.__storage
        else:
            return None

    @storage.setter
    def storage(self, value):
        self.__storage = value

    @property
    def storage_valid(self):
        return self.__storage_valid

    @storage_valid.setter
    def storage_valid(self, value):
        self.__storage_valid = value

    @property
    def range(self):
        """ tuple: Register range (min, max), None if undefined. """
        if self.__range:
            return (self.__range[0], self.__range[1])
        return None

    @property
    def labels(self):
        """ LabelsDictionary: Labels dictionary. """
        return self.__labels

    @property
    def enums(self):
        """ Enumerations list. """
        return self.__enums

    @property
    def enums_count(self):
        """ int: Register Enumerations count. """
        return self.__enums_count

    @property
    def cat_id(self):
        """Category ID."""
        return self.__cat_id

    @property
    def scat_id(self):
        """Sub-category ID."""
        return self.__scat_id

    @property
    def internal_use(self):
        """ int: Register internal_use. """
        return self.__internal_use



from enum import Enum

from ._ingenialink import ffi, lib
from ._utils import cstr, pstr, INT_SIZES

from .dict_labels import LabelsDictionary

# CANOPEN DTYPES
IL_REG_DTYPE_DOMAIN = 15


class REG_DTYPE(Enum):
    """
    Data Type.
    """
    U8 = lib.IL_REG_DTYPE_U8
    """ Unsigned 8-bit integer. """
    S8 = lib.IL_REG_DTYPE_S8
    """ Signed 8-bit integer. """
    U16 = lib.IL_REG_DTYPE_U16
    """ Unsigned 16-bit integer. """
    S16 = lib.IL_REG_DTYPE_S16
    """ Signed 16-bit integer. """
    U32 = lib.IL_REG_DTYPE_U32
    """ Unsigned 32-bit integer. """
    S32 = lib.IL_REG_DTYPE_S32
    """ Signed 32-bit integer. """
    U64 = lib.IL_REG_DTYPE_U64
    """ Unsigned 64-bit integer. """
    S64 = lib.IL_REG_DTYPE_S64
    """ Signed 64-bit integer. """
    FLOAT = lib.IL_REG_DTYPE_FLOAT
    """ Float. """
    STR = lib.IL_REG_DTYPE_STR
    """ String. """
    DOMAIN = IL_REG_DTYPE_DOMAIN
    """ Domain. """


class REG_ACCESS(Enum):
    """
    Access Type.
    """
    RW = lib.IL_REG_ACCESS_RW
    """ Read/Write. """
    RO = lib.IL_REG_ACCESS_RO
    """ Read-only. """
    WO = lib.IL_REG_ACCESS_WO
    """ Write-only. """


class REG_PHY(Enum):
    """
    Physical Units.
    """
    NONE = lib.IL_REG_PHY_NONE
    """ None. """
    TORQUE = lib.IL_REG_PHY_TORQUE
    """ Torque. """
    POS = lib.IL_REG_PHY_POS
    """ Position. """
    VEL = lib.IL_REG_PHY_VEL
    """ Velocity. """
    ACC = lib.IL_REG_PHY_ACC
    """ Acceleration. """
    VOLT_REL = lib.IL_REG_PHY_VOLT_REL
    """ Relative voltage (DC). """
    RAD = lib.IL_REG_PHY_RAD
    """ Radians. """


def _get_reg_id(reg, subnode=1):
    """
    Obtain Register and ID.

    Args:
        reg (str, Register): Register.
    """
    if isinstance(reg, str):
        return ffi.NULL, cstr(reg)
    elif isinstance(reg, Register):
        return reg._reg, ffi.NULL

    raise TypeError('Unexpected register type')


class Register(object):
    """
    Register.

    Args:
        identifier (str): Identifier.
        units (str): Units.
        address (int): Address.
        dtype (REG_DTYPE): Data type.
        access (REG_ACCESS): Access type.
        phy (REG_PHY, optional): Physical units.
        subnode (int): Subnode
        storage (any, optional): Storage.
        range (tuple, optional): Range (min, max).
        labels (dict, optional): Register labels.
        enums (dict, optional): Enumeration values.
        cat_id (str, optional): Category ID.
        scat_id (str, optional): Sub-category ID.
        internal_use (int, optional): Internal use.

    Raises:
        TypeError: If any of the parameters has invalid type.
    """
    def __init__(self, identifier, units, cyclic, address, dtype, access,
                 phy=REG_PHY.NONE, subnode=1, storage=None, range=None,
                 labels={}, enums=[], enums_count=0, cat_id=None, scat_id=None,
                 internal_use=0):
        if not isinstance(dtype, REG_DTYPE):
            raise TypeError('Invalid data type')

        if not isinstance(access, REG_ACCESS):
            raise TypeError('Invalid access type')

        if not isinstance(phy, REG_PHY):
            raise TypeError('Invalid physical units type')

        self._reg = ffi.new('il_reg_t *')

        # initialize register
        self._reg.identifier = ffi.new("char[]", cstr(identifier))
        self._reg.units = ffi.new("char[]", cstr(units))
        self._reg.address = address
        self._reg.subnode = subnode
        self._reg.cyclic = ffi.new("char[]", cstr(cyclic))
        self._reg.dtype = dtype.value
        self._reg.access = access.value
        self._reg.phy = phy.value
        self._reg.internal_use = internal_use

        self._reg.storage_valid = 0 if not storage else 1

        if dtype == REG_DTYPE.S8:
            if storage:
                self._reg.storage.s8 = int(storage)

            self._reg.range.min.s8 = (range[0] if range else
                                      INT_SIZES.S8_MIN.value)
            self._reg.range.max.s8 = (range[1] if range else
                                      INT_SIZES.S8_MAX.value)
        elif dtype == REG_DTYPE.U8:
            if storage:
                self._reg.storage.u8 = int(storage)

            self._reg.range.min.u8 = range[0] if range else 0
            self._reg.range.max.u8 = (range[1] if range else
                                      INT_SIZES.U8_MAX.value)
        if dtype == REG_DTYPE.S16:
            if storage:
                self._reg.storage.s16 = int(storage)

            self._reg.range.min.s16 = (range[0] if range else
                                       INT_SIZES.S16_MIN.value)
            self._reg.range.max.s16 = (range[1] if range else
                                       INT_SIZES.S16_MAX.value)
        elif dtype == REG_DTYPE.U16:
            if storage:
                self._reg.storage.u16 = int(storage)

            self._reg.range.min.u16 = range[0] if range else 0
            self._reg.range.max.u16 = (range[1] if range else
                                       INT_SIZES.U16_MAX.value)
        if dtype == REG_DTYPE.S32:
            if storage:
                self._reg.storage.s32 = int(storage)

            self._reg.range.min.s32 = (range[0] if range else
                                       INT_SIZES.S32_MIN.value)
            self._reg.range.max.s32 = (range[1] if range else
                                       INT_SIZES.S32_MAX.value)
        elif dtype == REG_DTYPE.U32:
            if storage:
                self._reg.storage.u32 = int(storage)

            self._reg.range.min.u32 = range[0] if range else 0
            self._reg.range.max.u32 = (range[1] if range else
                                       INT_SIZES.U32_MAX.value)
        if dtype == REG_DTYPE.S64:
            if storage:
                self._reg.storage.s64 = int(storage)

            self._reg.range.min.s64 = (range[0] if range else
                                       INT_SIZES.S64_MIN.value)
            self._reg.range.max.s64 = (range[1] if range else
                                       INT_SIZES.S64_MAX.value)
        elif dtype == REG_DTYPE.U64:
            if storage:
                self._reg.storage.u64 = int(storage)

            self._reg.range.min.u64 = range[0] if range else 0
            self._reg.range.max.u64 = (range[1] if range else
                                       INT_SIZES.U64_MAX.value)
        elif dtype == REG_DTYPE.FLOAT:
            if storage:
                self._reg.storage.flt = float(storage)

            self._reg.range.min.flt = (range[0] if range else -2147483648)
            self._reg.range.max.flt = (range[1] if range else 2147483647)
        else:
            self._reg.storage_valid = 0

        self._labels = LabelsDictionary(labels)
        self._reg.labels = self._labels._labels
        self._reg.enums_count = enums_count

        self._reg.cat_id = ffi.NULL if not cat_id else cstr(cat_id)

        if not cat_id and scat_id:
            raise ValueError('Sub-category requires a parent category')

        self._reg.scat_id = ffi.NULL if not scat_id else cstr(scat_id)

    def __repr__(self):
        """
        Obtain register object.

        Returns:
            str: Register information.
        """
        # obtain category/subcategory information
        if self.cat_id:
            cat_info = self.cat_id
            if self.scat_id:
                cat_info += ' + ' + self.scat_id
        else:
            cat_info = 'Uncategorized'

        if self.storage and self.storage_valid:
            storage_info = self.storage
        else:
            storage_info = 'No storage'

        return '<Register: {}, {}, {}, {}, 0x{:08x}, {}{}, {}, {}, [], {},' \
               'ST: {}, [{}], {}>'.format(
                    self.identifier,
                    self.units,
                    self.subnode,
                    self.cyclic,
                    self.address,
                    self.dtype,
                    ' ∊ ' + str(self.range) if self.range else '',
                    self.access,
                    self.phy,
                    self.enums,
                    self.enums_count,
                    storage_info,
                    cat_info,
                    self.internal_use
                )

    @classmethod
    def _from_register(cls, reg):
        """
        Create a new class instance from an existing register.

        Args:
            reg (Register): Register instance.

        Returns:
            Register: Register object instance copied.
        """
        inst = cls.__new__(cls)
        inst._reg = reg
        inst._labels = LabelsDictionary._from_labels(reg.labels)

        return inst

    @property
    def identifier(self):
        """
        Obtain register identifier.

        Returns:
            str: Register identifier
        """
        if self._reg.identifier != ffi.NULL:
            return pstr(self._reg.identifier)

        return None

    @property
    def units(self):
        """
        Obtain register units.

        Returns:
            str: Register units
        """
        if self._reg.units != ffi.NULL:
            return pstr(self._reg.units)

    @property
    def address(self):
        """
        Obtain register address.

        Returns:
            int: Register address.
        """
        return self._reg.address

    @property
    def subnode(self):
        """
        Obtain register subnode.

        Returns:
            int: Register subnode.
        """
        return self._reg.subnode

    @property
    def cyclic(self):
        """
        Obtain register cyclic.

        Returns:
            str: Register cyclic type.
        """
        if self._reg.cyclic != ffi.NULL:
            return pstr(self._reg.cyclic)

    @property
    def dtype(self):
        """
        Obtain register dtype.

        Returns:
            int: Register data type.
        """
        return REG_DTYPE(self._reg.dtype)

    @property
    def access(self):
        """
        Obtain register access.

        Returns:
            int: Register access type.
        """
        return REG_ACCESS(self._reg.access)

    @property
    def phy(self):
        """
        Obtain register physical units.

        Returns:
            int: Register physical units.
        """
        return REG_PHY(self._reg.phy)

    @property
    def storage(self):
        """
        Obtain register storage.

        Returns:
             int: Register storage.
        """
        if not self._reg.storage_valid:
            return None

        if self.dtype == REG_DTYPE.S8:
            return self._reg.storage.s8
        elif self.dtype == REG_DTYPE.U8:
            return self._reg.storage.u8
        if self.dtype == REG_DTYPE.S16:
            return self._reg.storage.s16
        elif self.dtype == REG_DTYPE.U16:
            return self._reg.storage.u16
        if self.dtype == REG_DTYPE.S32:
            return self._reg.storage.s32
        elif self.dtype == REG_DTYPE.U32:
            return self._reg.storage.u32
        if self.dtype == REG_DTYPE.S64:
            return self._reg.storage.s64
        elif self.dtype == REG_DTYPE.U64:
            return self._reg.storage.u64
        elif self.dtype == REG_DTYPE.FLOAT:
            return self._reg.storage.flt
        else:
            return None

    @property
    def range(self):
        """
        Obtains register range.

        Returns:
            tuple: Register range (min, max), None if undefined.
        """
        if self.dtype == REG_DTYPE.S8:
            return (self._reg.range.min.s8, self._reg.range.max.s8)
        elif self.dtype == REG_DTYPE.U8:
            return (self._reg.range.min.u8, self._reg.range.max.u8)
        if self.dtype == REG_DTYPE.S16:
            return (self._reg.range.min.s16, self._reg.range.max.s16)
        elif self.dtype == REG_DTYPE.U16:
            return (self._reg.range.min.u16, self._reg.range.max.u16)
        if self.dtype == REG_DTYPE.S32:
            return (self._reg.range.min.s32, self._reg.range.max.s32)
        elif self.dtype == REG_DTYPE.U32:
            return (self._reg.range.min.u32, self._reg.range.max.u32)
        if self.dtype == REG_DTYPE.S64:
            return (self._reg.range.min.s64, self._reg.range.max.s64)
        elif self.dtype == REG_DTYPE.U64:
            return (self._reg.range.min.u64, self._reg.range.max.u64)
        elif self.dtype == REG_DTYPE.FLOAT:
            return (self._reg.range.min.flt, self._reg.range.max.flt)

        return None

    @property
    def labels(self):
        """
        Obtains register labels.

        Returns:
            LabelsDictionary: Labels dictionary.
        """
        return self._labels

    @property
    def enums(self):
        """
        Obtain enumerations list of the register.

        Returns:
            array: Enumerations of the register.
        """
        if not hasattr(self, '_enums'):
            self._enums = []
            for i in range(0, self._reg.enums_count):
                dict = {
                    'label': pstr(self._reg.enums[i].label),
                    'value': self._reg.enums[i].value
                }
                self._enums.append(dict)
        return self._enums

    @property
    def enums_count(self):
        """
        Obtain number of enumerations of the register.

        Returns:
            int: Register Enumerations count.
        """
        return self._reg.enums_count

    @property
    def cat_id(self):
        """
        Category identifier.

        Returns:
            str | None: Current category identifier.
        """
        if self._reg.cat_id != ffi.NULL:
            return pstr(self._reg.cat_id)

        return None

    @property
    def scat_id(self):
        """
        Subcategory identifier.

        Returns:
            str | None: Current subcategory identifier.
        """
        if self._reg.scat_id != ffi.NULL:
            return pstr(self._reg.scat_id)
        return None

    @property
    def internal_use(self):
        """
        Internal use check.

        Returns:
            int: Register internal_use.
        """
        return self._reg.internal_use



from enum import Enum
from time import sleep

from ._ingenialink import lib, ffi
from ._utils import cstr, pstr, raise_null, raise_err, to_ms
from .registers import REG_DTYPE
from .exceptions import *

import numpy as np
import socket
import struct
import binascii
import os


CMD_CHANGE_CPU = 0x67E4


class NET_PROT(Enum):
    """
    Network Protocol.
    """
    EUSB = lib.IL_NET_PROT_EUSB
    """ E-USB. """
    MCB = lib.IL_NET_PROT_MCB
    """ MCB. """
    ETH = lib.IL_NET_PROT_ETH
    """ ETH. """
    ECAT = lib.IL_NET_PROT_ECAT
    """ ECAT. """
    CAN = 5
    """ CAN. """


class NET_STATE(Enum):
    """
    Network State.
    """
    CONNECTED = lib.IL_NET_STATE_CONNECTED
    """ Connected. """
    DISCONNECTED = lib.IL_NET_STATE_DISCONNECTED
    """ Disconnected. """
    FAULTY = lib.IL_NET_STATE_FAULTY
    """ Faulty. """


class NET_DEV_EVT(Enum):
    """
    Device Event.
    """
    ADDED = lib.IL_NET_DEV_EVT_ADDED
    """ Added. """
    REMOVED = lib.IL_NET_DEV_EVT_REMOVED
    """ Event. """


class EEPROM_TOOL_MODE(Enum):
    """
    EEPROM tool mode.
    """
    MODE_NONE = 0
    """ None. """
    MODE_READBIN = 1
    """ Read Binary. """
    MODE_READINTEL = 2
    """ Read Intelhex. """
    MODE_WRITEBIN = 3
    """ Write Binary. """
    MODE_WRITEINTEL = 4
    """ Write Intelhex. """
    MODE_WRITEALIAS = 5
    """ Write Alias. """
    MODE_INFO = 6
    """ Information. """


class NET_TRANS_PROT(Enum):
    """
    Protocol.
    """
    TCP = 1
    """ TCP. """
    UDP = 2
    """ UDP. """


def devices(prot):
    """
    Obtain a list of network devices.

    Args:
        prot (NET_PROT): Protocol.

    Returns:
        list: List of network devices.

    Raises:
        TypeError: If the protocol type is invalid.
    """
    if not isinstance(prot, NET_PROT):
        raise TypeError('Invalid protocol')

    devs = lib.il_net_dev_list_get(prot.value)

    found = []
    curr = devs

    while curr:
        found.append(pstr(curr.port))
        curr = curr.next

    lib.il_net_dev_list_destroy(devs)

    return found


def eeprom_tool(ifname, mode, filename):
    """
    Tool to modify and verify drive EEPROM.

    Args:
        ifname (str): Interface name.
        mode (int): EEPROM tool mode.
        filename (str): Path to the EEPROM file.

    Returns:
        int: Result code.

    """
    net__ = ffi.new('il_net_t **')
    ifname = cstr(ifname) if ifname else ffi.NULL
    filename = cstr(filename) if filename else ffi.NULL

    return lib.il_net_eeprom_tool(net__, ifname, 1, mode.value, filename)


def master_startup(ifname, if_address_ip):
    """
    Start SOEM master.

    Args:
        ifname (str): Interface name.
        if_address_ip (str): Interface address IP.

    Returns:
        int: Result code.
    """
    net__ = ffi.new('il_net_t **')
    ifname = cstr(ifname) if ifname else ffi.NULL
    if_address_ip = cstr(if_address_ip) if if_address_ip else ffi.NULL

    return lib.il_net_master_startup(net__, ifname, if_address_ip), net__


def num_slaves_get(ifname):
    ifname = cstr(ifname) if ifname else ffi.NULL
    return lib.il_net_num_slaves_get(ifname)


def master_stop(net):
    """
    Stop SOEM master.

    Returns:
        int: Result code.
    """
    return lib.il_net_master_stop(net)


def update_firmware_moco(node, subnode, ip, port, moco_file):
    """
    Update MOCO firmware through UDP protocol.

    Args:
        node: Network node.
        subnode: Drive subnode.
        ip: Drive address IP.
        port: Drive port.
        moco_file: Path to the firmware file.

    Returns:
        int: Result code.
    """
    r = 1
    upd = UDP(port, ip)

    if moco_file and os.path.isfile(moco_file):
        moco_in = open(moco_file, "r")

        print('Loading firmware...')
        try:
            for line in moco_in:
                words = line.split()

                # Get command and address
                cmd = int(words[1] + words[0], 16)
                data = b''
                data_start_byte = 2
                while data_start_byte in range(data_start_byte, len(words)):
                    # Load UDP data
                    data = data + bytes([int(words[data_start_byte], 16)])
                    data_start_byte = data_start_byte + 1

                # Send message
                upd.raw_cmd(node, subnode, cmd, data)

                if cmd == CMD_CHANGE_CPU:
                    sleep(1)

            print('Bootloading process succeeded!')
        except Exception as e:
            print('Error during bootloading process. {}'.format(e))
            r = -2
    else:
        print('File not found')
        r = -1

    return r


def update_firmware(ifname, filename, is_summit=False, slave=1):
    """
    Update firmware through FoE.

    Args:
        ifname: Interface name.
        filename: Path to the firmware file.
        is_summit:  [true] -> Everest
                    [false] -> Capitan or Low-Power drives
        slave: Slave number in the network.

    Returns:
        int: Result code.
    """
    net__ = ffi.new('il_net_t **')
    ifname = cstr(ifname) if ifname else ffi.NULL
    filename = cstr(filename) if filename else ffi.NULL
    return net__, lib.il_net_update_firmware(net__, ifname, slave,
                                             filename, is_summit)


def force_error(ifname, if_address_ip):
    """
    Force state machine error.

    Args:
        ifname: Interface name.
        if_address_ip: Interface address IP.

    Returns:
        int: Result code.
    """
    net__ = ffi.new('il_net_t **')
    ifname = cstr(ifname) if ifname else ffi.NULL
    if_address_ip = cstr(if_address_ip) if if_address_ip else ffi.NULL

    return lib.il_net_force_error(net__, ifname, if_address_ip)


@ffi.def_extern()
def _on_found_cb(ctx, servo_id):
    """
    On found callback shim.
    """
    self = ffi.from_handle(ctx)
    self._on_found(int(servo_id))


class UDP(object):
    def __init__(self, port, ip):
        self.port = port
        self.ip = ip
        self.rcv_buffer_size = 512
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, 0)
        self.socket.settimeout(8)
        self.socket.connect((ip, port))

    def __del__(self):
        try:
            self.socket.close()
        except Exception as e:
            print('Socket already closed. Exception: '.format(e))

    def close(self):
        self.socket.close()
        print('Socket closed')

    def write(self, frame):
        self.socket.sendto(frame, (self.ip, self.port))
        self.check_ack()

    def read(self):
        data, address = self.socket.recvfrom(self.rcv_buffer_size)
        return data

    def check_ack(self):
        rcv = self.read()
        ret_cmd = self.unmsg(rcv)
        if ret_cmd != 3:
            self.socket.close()
            raise Exception('No ACK received (command received %d)' % ret_cmd)
        return ret_cmd

    @staticmethod
    def unmsg(in_frame):
        # Base uart frame (subnode [4 bits], node [12 bits],
        # Addr [12 bits], cmd [3 bits], pending [1 bit],
        # Data [8 bytes]) and CRC [2 bytes] is 14 bytes long
        header = in_frame[2:4]
        cmd = struct.unpack('<H', header)[0] >> 1
        cmd = cmd & 0x7

        # CRC is computed with header and data (removing Tx CRC)
        crc = binascii.crc_hqx(in_frame[0:12], 0)
        crcread = struct.unpack('<H', in_frame[12:14])[0]
        if crcread != crc:
            raise UDPException('CRC error')

        return cmd

    @staticmethod
    def raw_msg(node, subnode, cmd, data, size):
        node_head = (node << 4) | (subnode & 0xf)
        node_head = struct.pack('<H', node_head)

        if size > 8:
            cmd = cmd + 1
            head = struct.pack('<H', cmd)
            head_size = struct.pack('<H', size)
            head_size = head_size + bytes([0] * (8 - len(head_size)))
            ret = node_head + head + head_size + struct.pack('<H',
                                                             binascii.crc_hqx(node_head + head + head_size, 0)) + data
        else:
            head = struct.pack('<H', cmd)
            ret = node_head + head + data + struct.pack('<H', binascii.crc_hqx(node_head + head + data, 0))

        return ret

    def raw_cmd(self, node, subnode, cmd, data):
        if len(data) > 8:
            frame = self.raw_msg(node, subnode, cmd, data, len(data))
        else:
            data = data + bytes([0] * (8 - len(data)))
            frame = self.raw_msg(node, subnode, cmd, data, len(data))

        self.write(frame)


class Network(object):
    """
    Network.

    Args:
        prot (NET_PROT): Protocol.
        port (str): Network device port (e.g. COM1, /dev/ttyACM0, etc.).
        timeout_rd (int, float, optional): Read timeout (s).
        timeout_wr (int, float, optional): Write timeout (s).

    Raises:
        TypeError: If the protocol type is invalid.
        ILCreationError: If the network cannot be created.
    """
    def __init__(self, prot, port=None, slave=1, timeout_rd=0.5, timeout_wr=0.5):
        if not isinstance(prot, NET_PROT):
            raise TypeError('Invalid protocol')

        if prot != NET_PROT.ECAT:

            port_ = ffi.new('char []', cstr(port))
            opts = ffi.new('il_net_opts_t *')

            opts.port = port_
            opts.timeout_rd = to_ms(timeout_rd)
            opts.timeout_wr = to_ms(timeout_wr)

            self._net = lib.il_net_create(prot.value, opts)
            raise_null(self._net)
        else:
            self.slave = slave
            self._net = ffi.new('il_net_t **')

    @classmethod
    def _from_existing(cls, net):
        """
        Create a new class instance from an existing network.

        Args:
            net (Network): Instance to copy.

        Returns:
            Network: New instanced class.

        """
        inst = cls.__new__(cls)
        inst._net = ffi.gc(net, lib.il_net_fake_destroy)

        return inst

    def master_startup(self, ifname, if_address_ip):
        """
        Start SOEM master.

        Args:
            ifname (str): Interface name.
            if_address_ip (str): Interface address IP.

        Returns:
            int: Result code.
        """
        ifname = cstr(ifname) if ifname else ffi.NULL
        if_address_ip = cstr(if_address_ip) if if_address_ip else ffi.NULL

        return lib.il_net_master_startup(self._net, ifname, if_address_ip)

    def set_if_params(self, ifname, if_address_ip):
        """
        Set ethernet interface parameters.

        Args:
            ifname (str): Interface name.
            if_address_ip (str): Interface address IP.

        Returns:

        """
        return lib.il_net_set_if_params(self._net, ifname, if_address_ip)

    def master_stop(self):
        """
        Stop SOEM master.

        Returns:
            int: Result code.
        """
        return lib.il_net_master_stop(self._net)

    def monitoring_channel_data(self, channel, dtype):
        """
        Obtain processed monitoring data of a channel.

        Args:
            channel (int): Identity channel number.
            dtype (REG_DTYPE): Data type of the register to map.

        Returns:
            array: Monitoring data.
        """
        data_arr = []
        size = int(self.monitoring_data_size)
        bytes_per_block = self.monitoring_get_bytes_per_block()
        if dtype == REG_DTYPE.U16:
            data_arr = lib.il_net_monitoring_channel_u16(self._net, channel)
        elif dtype == REG_DTYPE.S16:
            data_arr = lib.il_net_monitoring_channel_s16(self._net, channel)
        elif dtype == REG_DTYPE.U32:
            data_arr = lib.il_net_monitoring_channel_u32(self._net, channel)
        elif dtype == REG_DTYPE.S32:
            data_arr = lib.il_net_monitoring_channel_s32(self._net, channel)
        elif dtype == REG_DTYPE.FLOAT:
            data_arr = lib.il_net_monitoring_channel_flt(self._net, channel)
        ret_arr = []
        for i in range(0, int(size / bytes_per_block)):
            ret_arr.append(data_arr[i])
        return ret_arr

    def monitoring_remove_all_mapped_registers(self):
        """
        Remove all monitoring mapped registers.

        Returns:
            int: Result code.
        """
        return lib.il_net_remove_all_mapped_registers(self._net)

    def monitoring_set_mapped_register(self, channel, reg_idx, dtype):
        """
        Set monitoring mapped register.

        Args:
            channel (int): Identity channel number.
            reg_idx (int): Register address to map.
            dtype (REG_DTYPE): Data type of the register to map.

        Returns:
            int: Result code.
        """
        return lib.il_net_set_mapped_register(self._net, channel,
                                              reg_idx, dtype)

    def monitoring_get_num_mapped_registers(self):
        """
        Obtain the number of mapped registers.

        Returns:
            int: Actual number of mapped registers.
        """
        return lib.il_net_num_mapped_registers_get(self._net)

    def monitoring_enable(self):
        """
        Enable monitoring process.

        Returns:
            int: Result code.
        """
        return lib.il_net_enable_monitoring(self._net)

    def monitoring_disable(self):
        """
        Disable monitoring process.

        Returns:
            int: Result code.
        """
        return lib.il_net_disable_monitoring(self._net)

    def monitoring_read_data(self):
        """
        Obtain processed monitoring data.

        Returns:
            array: Actual processed monitoring data.
        """
        return lib.il_net_read_monitoring_data(self._net)

    def monitoring_get_bytes_per_block(self):
        """
        Obtain Bytes x Block configured.

        Returns:
            int: Actual number of Bytes x Block configured.
        """
        return lib.il_net_monitornig_bytes_per_block_get(self._net)

    # Disturbance
    def disturbance_channel_data(self, channel, dtype, data_arr):
        """
        Send disturbance data.

        Args:
            channel (int): Identity channel number.
            dtype (REG_DTYPE): Data type of the register mapped.
            data_arr (array): Data that will be sent to the drive.

        Returns:
            int: Return code.

        """
        if dtype == REG_DTYPE.U16:
            lib.il_net_disturbance_data_u16_set(self._net, channel, data_arr)
        elif dtype == REG_DTYPE.S16:
            lib.il_net_disturbance_data_s16_set(self._net, channel, data_arr)
        elif dtype == REG_DTYPE.U32:
            lib.il_net_disturbance_data_u32_set(self._net, channel, data_arr)
        elif dtype == REG_DTYPE.S32:
            lib.il_net_disturbance_data_s32_set(self._net, channel, data_arr)
        elif dtype == REG_DTYPE.FLOAT:
            lib.il_net_disturbance_data_flt_set(self._net, channel, data_arr)
        return 0

    def disturbance_remove_all_mapped_registers(self):
        """
        Remove all disturbance mapped registers.

        Returns:
            int: Return code.
        """
        return lib.il_net_disturbance_remove_all_mapped_registers(self._net)

    def disturbance_set_mapped_register(self, channel, address, dtype):
        """
        Set disturbance mapped register.

        Args:
            channel (int): Identity channel number.
            address (int): Register address to map.
            dtype (REG_DTYPE): Data type of the register to map.

        Returns:
            int: Return code.
        """
        return lib.il_net_disturbance_set_mapped_register(self._net, channel,
                                                          address, dtype)

    # Properties
    @property
    def prot(self):
        """
        Obtain network protocol.

        Returns:
            str: Current network protocol used.
        """
        return NET_PROT(lib.il_net_prot_get(self._net))

    @property
    def state(self):
        """
        Obtain network state.

        Returns:
            str: Current network state.
        """
        return NET_STATE(lib.il_net_state_get(self._net))

    @property
    def status(self):
        """
        Obtain network status.

        Returns:
            str: Current network status.
        """
        return lib.il_net_status_get(self._net)

    @property
    def port(self):
        """
        Obtain network port.

        Returns:
            str: Current network port.
        """
        port = lib.il_net_port_get(self._net)
        return pstr(port)

    @property
    def extended_buffer(self):
        """
        Obtain extended buffer data.

        Returns:
            str: Current extended buffer data.
        """
        ext_buff = lib.il_net_extended_buffer_get(self._net)
        return pstr(ext_buff)

    @property
    def monitoring_data(self):
        """
        Obtain monitoring data.

        Returns:
            array: Current monitoring data.
        """
        monitoring_data = lib.il_net_monitornig_data_get(self._net)
        size = int(self.monitoring_data_size / 2)
        ret_arr = []
        for i in range(0, size):
            ret_arr.append(monitoring_data[i])
        return ret_arr

    @property
    def monitoring_data_size(self):
        """
        Obtain monitoring data size.

        Returns:
            int: Current monitoring data size.
        """
        return lib.il_net_monitornig_data_size_get(self._net)

    @property
    def disturbance_data(self):
        """
        Obtain disturbance data.

        Returns:
            array: Current disturbance data.
        """
        disturbance_data = lib.il_net_disturbance_data_get(self._net)
        size = int(self.disturbance_data_size / 2)
        ret_arr = []
        for i in range(0, size):
            ret_arr.append(disturbance_data[i])
        return ret_arr

    @disturbance_data.setter
    def disturbance_data(self, value):
        """
        Set disturbance data.

        Args:
            value (array): Array with the disturbance to send.
        """
        disturbance_arr = value
        disturbance_arr = \
            np.pad(disturbance_arr,
                   (0, int(self.disturbance_data_size / 2) - len(value)),
                   'constant')
        lib.il_net_disturbance_data_set(self._net, disturbance_arr.tolist())

    @property
    def disturbance_data_size(self):
        """
        Obtain disturbance data size.

        Returns:
            int: Current disturbance data size.
        """
        return lib.il_net_disturbance_data_size_get(self._net)

    @disturbance_data_size.setter
    def disturbance_data_size(self, value):
        """
        Set disturbance data size.

        Args:
            value (int): Disturbance data size in bytes.
        """
        lib.il_net_disturbance_data_size_set(self._net, value)

    def close_socket(self):
        return lib.il_net_close_socket(self._net)

    def connect(self):
        """
        Connect network.
        """
        r = lib.il_net_connect(self._net)
        raise_err(r)

    def disconnect(self):
        """
        Disconnect network.
        """
        lib.il_net_disconnect(self._net)

    def servos(self, on_found=None):
        """
        Obtain a list of attached servos.

        Args:
            on_found (callback, optional): Servo found callback.

        Returns:
            list: List of attached servos.
        """
        if on_found:
            self._on_found = on_found

            callback = lib._on_found_cb
            handle = ffi.new_handle(self)
        else:
            self._on_found = ffi.NULL

            callback = ffi.NULL
            handle = ffi.NULL

        servos = lib.il_net_servos_list_get(self._net, callback, handle)

        found = []
        curr = servos

        while curr:
            found.append(int(curr.id))
            curr = curr.next

        lib.il_net_servos_list_destroy(servos)

        return found

    def net_mon_status(self, on_evt):
        """
        Calls given function everytime a connection/disconnection event is
        raised.

        Args:
            on_evt (Callback): Function that will be called every time an event
                            is raised.
        """
        if self.prot == NET_PROT.ETH or self.prot == NET_PROT.ECAT:
            status = self.status
            while True:
                if status != self.status:
                    if self.status == 0:
                        on_evt(NET_DEV_EVT.ADDED)
                    elif self.status == 1:
                        on_evt(NET_DEV_EVT.REMOVED)
                    status = self.status
                sleep(1)

    def net_mon_stop(self):
        """
        Stop monitoring network events.
        """
        lib.il_net_mon_stop(self._net)

    def destroy_network(self):
        """
        Destroy network instance.
        """
        lib.il_net_destroy(self._net)

    def set_reconnection_retries(self, retries):
        """
        Set the number of reconnection retries in our application.

        Args:
            retries (int): Number of reconnection retries.
        """
        return lib.il_net_set_reconnection_retries(self._net, retries)

    def set_recv_timeout(self, timeout):
        """
        Set receive communications timeout.

        Args:
            timeout (int): Timeout in ms.
        Returns:
            int: Result code.
        """
        return lib.il_net_set_recv_timeout(self._net, timeout)

    def set_status_check_stop(self, stop):
        """
        Start/Stop the internal monitor of the drive status.

        Args:
            stop (int): 0 to START, 1 to STOP.
        Returns:
            int: Result code.
        """
        return lib.il_net_set_status_check_stop(self._net, stop)


@ffi.def_extern()
def _on_evt_cb(ctx, evt, port):
    """
    On event callback shim.
    """
    self = ffi.from_handle(ctx)
    self._on_evt(NET_DEV_EVT(evt), pstr(port))


class NetworkMonitor(object):
    """
    Network Monitor.

    Args:
        prot (NET_PROT): Protocol.

    Raises:
        TypeError: If the protocol type is invalid.
        ILCreationError: If the monitor cannot be created.
    """
    def __init__(self, prot):
        if not isinstance(prot, NET_PROT):
            raise TypeError('Invalid protocol')

        mon = lib.il_net_dev_mon_create(prot.value)
        raise_null(mon)

        self._mon = ffi.gc(mon, lib.il_net_dev_mon_destroy)

    def start(self, on_evt):
        """
        Start the monitor.

        Args:
            on_evt (callback): Callback function.
        """
        self._on_evt = on_evt
        self._handle = ffi.new_handle(self)

        r = lib.il_net_dev_mon_start(self._mon, lib._on_evt_cb, self._handle)
        raise_err(r)

    def stop(self):
        """
        Stop the monitor.
        """
        lib.il_net_dev_mon_stop(self._mon)



import collections

from ._ingenialink import ffi, lib
from ._utils import cstr, pstr, raise_null, raise_err


class LabelsDictionary(collections.MutableMapping):
    """
    Labels dictionary.

    Args:
        labels (dict, optional): Labels.

    Raises:
        ILCreationError: If the dictionary could not be created.
    """
    def __init__(self, labels={}):
        _labels = lib.il_dict_labels_create()
        raise_null(_labels)

        self._labels = ffi.gc(_labels, lib.il_dict_labels_destroy)

        for lang, content in labels.items():
            lib.il_dict_labels_set(self._labels, cstr(lang), cstr(content))

        self._load_langs()

    @classmethod
    def _from_labels(cls, _labels):
        """
        Create a new class instance from an existing labels dictionary.
        """
        inst = cls.__new__(cls)
        inst._labels = _labels

        inst._load_langs()

        return inst

    def _load_langs(self):
        """
        Load languages from dictionary (cache).
        """
        langs = lib.il_dict_labels_langs_get(self._labels)

        self._langs = []
        i = 0
        lang = langs[0]
        while lang != ffi.NULL:
            self._langs.append(pstr(lang))
            i += 1
            lang = langs[i]

        lib.il_dict_labels_langs_destroy(langs)

    def __getitem__(self, lang):
        content_p = ffi.new('char **')
        r = lib.il_dict_labels_get(self._labels, cstr(lang), content_p)
        raise_err(r)

        return pstr(content_p[0])

    def __setitem__(self, lang, content):
        lib.il_dict_labels_set(self._labels, cstr(lang), cstr(content))
        self._langs.append(lang)

    def __delitem__(self, lang):
        lib.il_dict_labels_del(self._labels, cstr(lang))
        self._langs.remove(lang)

    def __len__(self):
        return len(self._langs)

    def __iter__(self):
        return iter(self._langs)



import time
import threading
import canopen
import struct
import xml.etree.ElementTree as ET

from .._utils import *
from .constants import *
from ..servo import SERVO_STATE
from .._ingenialink import ffi, lib
from .dictionary import DictionaryCANOpen
from .registers import Register, REG_DTYPE, REG_ACCESS


SERIAL_NUMBER = Register(
    identifier='', units='', subnode=1, idx="0x26E6", subidx="0x00",
    cyclic='CONFIG', dtype=REG_DTYPE.U32, access=REG_ACCESS.RO
)
PRODUCT_CODE = Register(
    identifier='', units='', subnode=1, idx="0x26E1", subidx="0x00",
    cyclic='CONFIG', dtype=REG_DTYPE.U32, access=REG_ACCESS.RO
)
SOFTWARE_VERSION = Register(
    identifier='', units='', subnode=1, idx="0x26E4", subidx="0x00",
    cyclic='CONFIG', dtype=REG_DTYPE.STR, access=REG_ACCESS.RO
)
REVISION_NUMBER = Register(
    identifier='', units='', subnode=1, idx="0x26E2", subidx="0x00",
    cyclic='CONFIG', dtype=REG_DTYPE.U32, access=REG_ACCESS.RO
)

STATUS_WORD_REGISTERS = {
    1: Register(
        identifier='', units='', subnode=1, idx="0x6041", subidx="0x00",
        cyclic='CYCLIC_TX', dtype=REG_DTYPE.U16, access=REG_ACCESS.RO
    ),
    2: Register(
        identifier='', units='', subnode=2, idx="0x6841", subidx="0x00",
        cyclic='CYCLIC_TX', dtype=REG_DTYPE.U16, access=REG_ACCESS.RO
    ),
    3: Register(
        identifier='', units='', subnode=3, idx="0x7041", subidx="0x00",
        cyclic='CYCLIC_TX', dtype=REG_DTYPE.U16, access=REG_ACCESS.RO
    )
}

CONTROL_WORD_REGISTERS = {
    1: Register(
        identifier='', units='', subnode=1, idx="0x2010", subidx="0x00",
        cyclic='CYCLIC_RX', dtype=REG_DTYPE.U16, access=REG_ACCESS.RW
    ),
    2: Register(
        identifier='', units='', subnode=2, idx="0x2810", subidx="0x00",
        cyclic='CYCLIC_RX', dtype=REG_DTYPE.U16, access=REG_ACCESS.RW
    ),
    3: Register(
        identifier='', units='', subnode=3, idx="0x3010", subidx="0x00",
        cyclic='CYCLIC_RX', dtype=REG_DTYPE.U16, access=REG_ACCESS.RW
    )
}

STORE_ALL_REGISTERS = {
    1: Register(
        identifier='', units='', subnode=1, idx="0x26DB", subidx="0x00",
        cyclic='CONFIG', dtype=REG_DTYPE.U32, access=REG_ACCESS.RW
    ),
    2: Register(
        identifier='', units='', subnode=2, idx="0x2EDB", subidx="0x00",
        cyclic='CONFIG', dtype=REG_DTYPE.U32, access=REG_ACCESS.RW
    ),
    3: Register(
        identifier='', units='', subnode=3, idx="0x36DB", subidx="0x00",
        cyclic='CONFIG', dtype=REG_DTYPE.U32, access=REG_ACCESS.RW
    )
}


class DriveStatusThread(threading.Thread):
    def __init__(self, parent):
        """ Constructor, setting initial variables """
        super(DriveStatusThread, self).__init__()
        self.__parent = parent
        self.__stop = False

    def run(self):
        while not self.__stop:
            for subnode in range(1, self.__parent.subnodes):
                try:
                    status_word = self.__parent.raw_read(
                        STATUS_WORD_REGISTERS[subnode], subnode=subnode
                    )
                    state = self.__parent.status_word_decode(status_word)
                    self.__parent.set_state(state, subnode=subnode)
                except Exception as e:
                    print('IL: Error getting drive status. '
                          'Exception: {}'.format(e))
            time.sleep(1.5)

    def activate_stop_flag(self):
        self.__stop = True


class Servo(object):
    def __init__(self, net, node, dict, boot_mode=False):
        self.__net = net
        self.__node = node
        self.__dict = DictionaryCANOpen(dict)
        self.__info = {}
        self.__state = {
            1: lib.IL_SERVO_STATE_NRDY,
            2: lib.IL_SERVO_STATE_NRDY,
            3: lib.IL_SERVO_STATE_NRDY
        }
        self.__observers = []
        self.__lock = threading.RLock()
        self.__units_torque = None
        self.__units_pos = None
        self.__units_vel = None
        self.__units_acc = None
        self.__name = "Drive"
        self.__drive_status_thread = None
        if not boot_mode:
            self.init_info()

    def init_info(self):
        serial_number = self.raw_read(SERIAL_NUMBER)
        product_code = self.raw_read(PRODUCT_CODE)
        sw_version = self.raw_read(SOFTWARE_VERSION)
        revision_number = self.raw_read(REVISION_NUMBER)
        hw_variant = 'A'
        # Set the current state of servo
        status_word = self.raw_read(STATUS_WORD_REGISTERS[1])
        state = self.status_word_decode(status_word)
        self.set_state(state, 1)
        self.__info = {
            'serial': serial_number,
            'name': self.__name,
            'sw_version': sw_version,
            'hw_variant': hw_variant,
            'prod_code': product_code,
            'revision': revision_number
        }
        self.__drive_status_thread = DriveStatusThread(self)
        self.__drive_status_thread.start()

    def stop_drive_status_thread(self):
        if self.__drive_status_thread is not None and \
                self.__drive_status_thread.is_alive():
            self.__drive_status_thread.activate_stop_flag()
            self.__drive_status_thread.join()
            self.__drive_status_thread = None

    def emcy_subscribe(self, callback):
        pass

    def emcy_unsubscribe(self, callback):
        pass

    def get_reg(self, reg, subnode=1):
        if isinstance(reg, Register):
            _reg = reg
        elif isinstance(reg, str):
            _dict = self.__dict
            if not _dict:
                raise_err(lib.IL_EIO, 'No dictionary loaded')
            if reg not in _dict.regs[subnode]:
                raise_err(lib.IL_EWRONGREG, 'Invalid register')
            _reg = _dict.regs[subnode][reg]
        else:
            raise_err(lib.IL_EWRONGREG, 'Invalid register')
        return _reg

    def raw_read(self, reg, subnode=1):
        """ Raw read from servo.

            Args:
                reg (Register): Register.

            Returns:
                int: Otained value

            Raises:
                TypeError: If the register type is not valid.
        """
        _reg = self.get_reg(reg, subnode)

        access = _reg.access
        if access == REG_ACCESS.WO:
            raise_err(lib.IL_EACCESS, 'Register is Write-only')

        value = None
        dtype = _reg.dtype
        error_raised = None
        try:
            self.__lock.acquire()
            if dtype == REG_DTYPE.S8:
                value = int.from_bytes(
                    self.__node.sdo.upload(int(str(_reg.idx), 16),
                                           int(str(_reg.subidx), 16)),
                    "little",
                    signed=True
                )
            elif dtype == REG_DTYPE.S16:
                value = int.from_bytes(
                    self.__node.sdo.upload(int(str(_reg.idx), 16),
                                           int(str(_reg.subidx), 16)),
                    "little",
                    signed=True
                )
            elif dtype == REG_DTYPE.S32:
                value = int.from_bytes(
                    self.__node.sdo.upload(int(str(_reg.idx), 16),
                                           int(str(_reg.subidx), 16)),
                    "little",
                    signed=True
                )
            elif dtype == REG_DTYPE.FLOAT:
                [value] = struct.unpack('f',
                                        self.__node.sdo.upload(
                                            int(str(_reg.idx), 16),
                                            int(str(_reg.subidx), 16))
                                        )
            elif dtype == REG_DTYPE.STR:
                value = self.__node.sdo.upload(
                            int(str(_reg.idx), 16),
                            int(str(_reg.subidx), 16)
                        ).decode("utf-8")
            else:
                value = int.from_bytes(
                    self.__node.sdo.upload(int(str(_reg.idx), 16),
                                           int(str(_reg.subidx), 16)),
                    "little"
                )
        except Exception as e:
            print(_reg.identifier + " : " + str(e))
            error_raised = "Error reading {}".format(_reg.identifier)
        finally:
            self.__lock.release()

        if error_raised is not None:
            raise_err(lib.IL_EIO, error_raised)

        return value

    def read(self, reg, subnode=1):
        """ Read from servo.

            Args:
                reg (str, Register): Register.

            Returns:
                float: Otained value

            Raises:
                TypeError: If the register type is not valid.
        """
        return self.raw_read(reg, subnode=subnode)

    def change_sdo_timeout(self, value):
        self.__node.sdo.RESPONSE_TIMEOUT = value

    def write(self, reg, data, confirm=True, extended=0, subnode=1):
        return self.raw_write(reg, data, confirm=True,
                              extended=0, subnode=subnode)

    def raw_write(self, reg, data, confirm=True, extended=0, subnode=1):
        """ Raw write to servo.

            Args:
                reg (Register): Register.
                data (int): Data.
                confirm (bool, optional): Confirm write.
                extended (int, optional): Extended frame.

            Raises:
                TypeError: If any of the arguments type is not valid or
                    unsupported.
        """

        _reg = self.get_reg(reg, subnode)

        if _reg.access == REG_ACCESS.RO:
            raise_err(lib.IL_EACCESS, 'Register is Read-only')

        # auto cast floats if register is not float
        if _reg.dtype == REG_DTYPE.FLOAT:
            data = float(data)
        elif _reg.dtype == REG_DTYPE.DOMAIN:
            pass
        else:
            data = int(data)

        error_raised = None
        try:
            self.__lock.acquire()
            if _reg.dtype == REG_DTYPE.FLOAT:
                self.__node.sdo.download(int(str(_reg.idx), 16),
                                         int(str(_reg.subidx), 16),
                                         struct.pack('f', data))
            elif _reg.dtype == REG_DTYPE.DOMAIN:
                self.__node.sdo.download(int(str(_reg.idx), 16),
                                         int(str(_reg.subidx), 16), data)
            else:
                bytes_length = 2
                signed = False
                if _reg.dtype == REG_DTYPE.U8:
                    bytes_length = 1
                elif _reg.dtype == REG_DTYPE.S8:
                    bytes_length = 1
                    signed = True
                elif _reg.dtype == REG_DTYPE.U16:
                    bytes_length = 2
                elif _reg.dtype == REG_DTYPE.S16:
                    bytes_length = 2
                    signed = True
                elif _reg.dtype == REG_DTYPE.U32:
                    bytes_length = 4
                elif _reg.dtype == REG_DTYPE.S32:
                    bytes_length = 4
                    signed = True

                self.__node.sdo.download(int(str(_reg.idx), 16),
                                         int(str(_reg.subidx), 16),
                                         data.to_bytes(bytes_length,
                                                       byteorder='little',
                                                       signed=signed))
        except Exception as e:
            print(_reg.identifier + " : " + str(e))
            error_raised = "Error writing {}".format(_reg.identifier)
        finally:
            self.__lock.release()

        if error_raised is not None:
            raise_err(lib.IL_EIO, error_raised)

    def get_all_registers(self):
        for obj in self.__node.object_dictionary.values():
            print('0x%X: %s' % (obj.index, obj.name))
            if isinstance(obj, canopen.objectdictionary.Record):
                for subobj in obj.values():
                    print('  %d: %s' % (subobj.subindex, subobj.name))

    def dict_storage_read(self, new_path, subnode=0):
        """Read all dictionary registers content and put it to the dictionary
        storage."""

        with open(self.__dict.dict, 'r') as xml_file:
            tree = ET.parse(xml_file)
        root = tree.getroot()

        axis = tree.findall('*/Device/Axes/Axis')
        if axis:
            # Multiaxis
            registers = root.findall(
                './Body/Device/Axes/Axis/Registers/Register'
            )
        else:
            # Single axis
            registers = root.findall('./Body/Device/Registers/Register')

        registers_category = root.find('Body/Device/Registers')

        for register in registers:
            try:
                element_subnode = int(register.attrib['subnode'])
                if subnode == 0 or subnode == element_subnode:
                    if register.attrib['access'] == 'rw':
                        storage = self.raw_read(register.attrib['id'], subnode=element_subnode)
                        register.set('storage', str(storage))

                        # Update register object
                        reg = self.__dict.regs[element_subnode][register.attrib['id']]
                        reg.storage = storage
                        reg.storage_valid = 1
                else:
                    registers_category.remove(register)
            except BaseException as e:
                print("Exception during dict_storage_read, register " + register.attrib['id'] + ": ", str(e))

        image = xml_data.find('./DriveImage')
        if image is not None:
            xml_data.remove(image)

        tree.write(new_path)
        xml_file.close()

    def dict_storage_write(self, path, subnode=0):
        """Write current dictionary storage to the servo drive."""
        with open(path, 'r') as xml_file:
            tree = ET.parse(xml_file)
        root = tree.getroot()

        axis = tree.findall('*/Device/Axes/Axis')
        if axis:
            # Multiaxis
            registers = root.findall(
                './Body/Device/Axes/Axis/Registers/Register'
            )
        else:
            # Single axis
            registers = root.findall('./Body/Device/Registers/Register')

        for element in registers:
            try:
                if 'storage' in element.attrib and element.attrib['access'] == 'rw':
                    if subnode == 0 or subnode == int(element.attrib['subnode']):
                        self.raw_write(element.attrib['id'], float(element.attrib['storage']),
                                       subnode=int(element.attrib['subnode'])
                                       )
            except BaseException as e:
                print("Exception during dict_storage_write, register " +
                      element.attrib['id'] + ": ", str(e))

    def store_all(self, subnode=1):
        """ Store all servo current parameters to the NVM. """
        r = 0
        try:
            self.raw_write(STORE_ALL_REGISTERS[subnode], 0x65766173,
                           subnode=subnode)
        except Exception as e:
            r = -1
        return r

    def dict_load(self, dict_f):
        """ Load dictionary.

            Args:
                dict_f (str): Dictionary.
        """
        try:
            self.__dict = DictionaryCANOpen(dict_f)
        except Exception as e:
            print("Error loading a dictionary")

    def state_subscribe(self, cb):
        """ Subscribe to state changes.

            Args:
                cb: Callback

            Returns:
                int: Assigned slot.
        """
        r = len(self.__observers)
        self.__observers.append(cb)
        return r

    def status_word_decode(self, status_word):
        if (status_word & IL_MC_PDS_STA_NRTSO_MSK) == IL_MC_PDS_STA_NRTSO:
            state = lib.IL_SERVO_STATE_NRDY
        elif (status_word & IL_MC_PDS_STA_SOD_MSK) == IL_MC_PDS_STA_SOD:
            state = lib.IL_SERVO_STATE_DISABLED
        elif (status_word & IL_MC_PDS_STA_RTSO_MSK) == IL_MC_PDS_STA_RTSO:
            state = lib.IL_SERVO_STATE_RDY
        elif (status_word & IL_MC_PDS_STA_SO_MSK) == IL_MC_PDS_STA_SO:
            state = lib.IL_SERVO_STATE_ON
        elif (status_word & IL_MC_PDS_STA_OE_MSK) == IL_MC_PDS_STA_OE:
            state = lib.IL_SERVO_STATE_ENABLED
        elif (status_word & IL_MC_PDS_STA_QSA_MSK) == IL_MC_PDS_STA_QSA:
            state = lib.IL_SERVO_STATE_QSTOP
        elif (status_word & IL_MC_PDS_STA_FRA_MSK) == IL_MC_PDS_STA_FRA:
            state = lib.IL_SERVO_STATE_FAULTR
        elif (status_word & IL_MC_PDS_STA_F_MSK) == IL_MC_PDS_STA_F:
            state = lib.IL_SERVO_STATE_FAULT
        else:
            state = lib.IL_SERVO_STATE_NRDY
        return SERVO_STATE(state)

    def set_state(self, state, subnode):
        current_state = self.__state[subnode]
        if current_state != state:
            self.state[subnode] = state
            for callback in self.__observers:
                callback(state, None, subnode)

    def status_word_wait_change(self, status_word, timeout, subnode=1):
        r = 0
        start_time = int(round(time.time() * 1000))
        actual_status_word = self.raw_read(STATUS_WORD_REGISTERS[subnode],
                                           subnode=1)
        while actual_status_word == status_word:
            current_time = int(round(time.time() * 1000))
            time_diff = (current_time - start_time)
            if time_diff > timeout:
                r = lib.IL_ETIMEDOUT
                return r
            actual_status_word = self.raw_read(STATUS_WORD_REGISTERS[subnode],
                                               subnode=1)
        return r

    def fault_reset(self, subnode=1):
        r = 0
        retries = 0
        status_word = self.raw_read(STATUS_WORD_REGISTERS[subnode],
                                    subnode=subnode)
        state = self.status_word_decode(status_word)
        self.set_state(state, subnode)
        while self.state[subnode].value == lib.IL_SERVO_STATE_FAULT or \
                self.state[subnode].value == lib.IL_SERVO_STATE_FAULTR:
            # Check if faulty, if so try to reset (0->1)
            if retries == FAULT_RESET_RETRIES:
                return lib.IL_ESTATE

            status_word = self.raw_read(STATUS_WORD_REGISTERS[subnode],
                                        subnode=subnode)
            self.raw_write(CONTROL_WORD_REGISTERS[subnode], 0, subnode=subnode)
            self.raw_write(CONTROL_WORD_REGISTERS[subnode], IL_MC_CW_FR,
                           subnode=subnode)
            # Wait until statusword changes
            r = self.status_word_wait_change(status_word, PDS_TIMEOUT,
                                             subnode=1)
            if r < 0:
                return r
            retries += 1
        return r

    def enable(self, timeout=2000, subnode=1):
        """ Enable PDS. """
        r = 0

        status_word = self.raw_read(STATUS_WORD_REGISTERS[subnode],
                                    subnode=subnode)
        state = self.status_word_decode(status_word)
        self.set_state(state, subnode)

        # Try fault reset if faulty
        if self.state[subnode].value == lib.IL_SERVO_STATE_FAULT or \
                self.state[subnode].value == lib.IL_SERVO_STATE_FAULTR:
            r = self.fault_reset(subnode=subnode)
            if r < 0:
                return r

        while self.state[subnode].value != lib.IL_SERVO_STATE_ENABLED:
            status_word = self.raw_read(STATUS_WORD_REGISTERS[subnode],
                                        subnode=subnode)
            state = self.status_word_decode(status_word)
            self.set_state(state, subnode)
            if self.state[subnode].value != lib.IL_SERVO_STATE_ENABLED:
                # Check state and commandaction to reach enabled
                cmd = IL_MC_PDS_CMD_EO
                if self.state[subnode].value == lib.IL_SERVO_STATE_FAULT:
                    return lib.IL_ESTATE
                elif self.state[subnode].value == lib.IL_SERVO_STATE_NRDY:
                    cmd = IL_MC_PDS_CMD_DV
                elif self.state[subnode].value == lib.IL_SERVO_STATE_DISABLED:
                    cmd = IL_MC_PDS_CMD_SD
                elif self.state[subnode].value == lib.IL_SERVO_STATE_RDY:
                    cmd = IL_MC_PDS_CMD_SOEO

                self.raw_write(CONTROL_WORD_REGISTERS[subnode], cmd,
                               subnode=subnode)

                # Wait for state change
                r = self.status_word_wait_change(status_word, PDS_TIMEOUT,
                                                 subnode=1)
                if r < 0:
                    return r

                # Read the current status word
                status_word = self.raw_read(STATUS_WORD_REGISTERS[subnode],
                                            subnode=subnode)
                state = self.status_word_decode(status_word)
                self.set_state(state, subnode)
        raise_err(r)

    def disable(self, subnode=1):
        """ Disable PDS. """
        r = 0

        status_word = self.raw_read(STATUS_WORD_REGISTERS[subnode],
                                    subnode=subnode)
        state = self.status_word_decode(status_word)
        self.set_state(state, subnode)

        while self.state[subnode].value != lib.IL_SERVO_STATE_DISABLED:
            state = self.status_word_decode(status_word)
            self.set_state(state, subnode)

            if self.state[subnode].value == lib.IL_SERVO_STATE_FAULT or \
                    self.state[subnode].value == lib.IL_SERVO_STATE_FAULTR:
                # Try fault reset if faulty
                r = self.fault_reset(subnode=subnode)
                if r < 0:
                    return r
                status_word = self.raw_read(STATUS_WORD_REGISTERS[subnode],
                                            subnode=subnode)
                state = self.status_word_decode(status_word)
                self.set_state(state, subnode)
            elif self.state[subnode].value != lib.IL_SERVO_STATE_DISABLED:
                # Check state and command action to reach disabled
                self.raw_write(CONTROL_WORD_REGISTERS[subnode],
                               IL_MC_PDS_CMD_DV, subnode=subnode)

                # Wait until statusword changes
                r = self.status_word_wait_change(status_word, PDS_TIMEOUT,
                                                 subnode=1)
                if r < 0:
                    return r
                status_word = self.raw_read(STATUS_WORD_REGISTERS[subnode],
                                            subnode=subnode)
                state = self.status_word_decode(status_word)
                self.set_state(state, subnode)
        raise_err(r)

    def get_state(self, subnode=1):
        return self.__state[subnode], None
    
    @property
    def net(self):
        """ net: CANopen Network. """
        return self.__net

    @net.setter
    def net(self, net):
        self.__net = net

    @property
    def name(self):
        """ name: Drive name. """
        return self.__name

    @name.setter
    def name(self, new_name):
        self.__name = new_name

    @property
    def dict(self):
        """ Dictionary: Dictionary. """
        return self.__dict

    @property
    def node(self):
        """ int: Node. """
        return self.__node

    @property
    def errors(self):
        """ dict: Errors. """
        return self.__dict.errors.errors

    @property
    def info(self):
        """ dict: Servo information. """
        return self.__info

    @property
    def state(self):
        """ tuple: Servo state and state flags. """
        return self.__state

    @state.setter
    def state(self, new_state):
        self.__state = new_state

    @property
    def units_torque(self):
        """ SERVO_UNITS_TORQUE: Torque units. """
        return self.__units_torque

    @units_torque.setter
    def units_torque(self, units):
        self.__units_torque = units

    @property
    def units_pos(self):
        """ SERVO_UNITS_POS: Position units. """
        return self.__units_pos

    @units_pos.setter
    def units_pos(self, units):
        self.__units_pos = units

    @property
    def units_vel(self):
        """ SERVO_UNITS_VEL: Velocity units. """
        return self.__units_vel

    @units_vel.setter
    def units_vel(self, units):
        self.__units_vel = units

    @property
    def units_acc(self):
        """ SERVO_UNITS_ACC: Acceleration units. """
        return self.__units_acc

    @units_acc.setter
    def units_acc(self, units):
        self.__units_acc = units

    @property
    def subnodes(self):
        """ SUBNODES: Number of subnodes. """
        return self.__dict.subnodes



import time
import canopen
from enum import Enum
from threading import Thread
from time import sleep

from .servo_node import Servo
from ..net import NET_PROT, NET_STATE

import logging

log = logging.getLogger(__name__)


CAN_CHANNELS = {
    'kvaser': (0, 1),
    'pcan': ('PCAN_USBBUS1', 'PCAN_USBBUS2'),
    'ixxat': (0, 1)
}


class CAN_DEVICE(Enum):
    """ CAN Device. """
    KVASER = 'kvaser'
    """ Kvaser. """
    PCAN = 'pcan'
    """ Peak. """
    IXXAT = 'ixxat'
    """ Ixxat. """


class CAN_BAUDRATE(Enum):
    """ Baudrates. """
    Baudrate_1M = 1000000
    """ 1 Mbit/s """
    Baudrate_500K = 500000
    """ 500 Kbit/s """
    Baudrate_250K = 250000
    """ 250 Kbit/s """
    Baudrate_125K = 125000
    """ 125 Kbit/s """
    Baudrate_100K = 100000
    """ 100 Kbit/s """
    Baudrate_50K = 50000
    """ 50 Kbit/s """


class CAN_BIT_TIMMING(Enum):
    """ Baudrates. """
    Baudrate_1M = 0
    """ 1 Mbit/s """
    Baudrate_500K = 2
    """ 500 Kbit/s """
    Baudrate_250K = 3
    """ 250 Kbit/s """
    Baudrate_125K = 4
    """ 150 Kbit/s """
    Baudrate_100K = 5
    """ 100 Kbit/s """
    Baudrate_50K = 6
    """ 50 Kbit/s """


class HearbeatThread(Thread):
    def __init__(self, parent, node):
        """ Constructor, setting initial variables """
        super(HearbeatThread, self).__init__()
        self.__parent = parent
        self.__node = node
        self.__timestamp = self.__node.nmt.timestamp
        self.__state = NET_STATE.CONNECTED
        self.__stop = False

    def run(self):
        while not self.__stop:
            if self.__timestamp == self.__node.nmt.timestamp:
                if self.__state != NET_STATE.DISCONNECTED:
                    self.__parent.net_state = NET_STATE.DISCONNECTED
                    self.__state = NET_STATE.DISCONNECTED
                else:
                    self.__parent.reset_network()
            else:
                if self.__state != NET_STATE.CONNECTED:
                    self.__parent.net_state = NET_STATE.CONNECTED
                    self.__state = NET_STATE.CONNECTED
                self.__timestamp = self.__node.nmt.timestamp
            sleep(1.5)

    def activate_stop_flag(self):
        self.__stop = True


class Network(object):
    def __init__(self, device=None, channel=0,
                 baudrate=CAN_BAUDRATE.Baudrate_1M):
        self.__servos = []
        self.__device = device.value
        self.__channel = CAN_CHANNELS[self.__device][channel]
        self.__baudrate = baudrate.value
        self.__network = canopen.Network()
        self.__net_state = NET_STATE.DISCONNECTED
        self.__observers = []
        self.__eds = None
        self.__dict = None
        self.__heartbeat_thread = None
        if device is not None:
            try:
                self.__network.connect(bustype=self.__device,
                                       channel=self.__channel,
                                       bitrate=self.__baudrate)
            except Exception as e:
                print('Exception trying to connect: ', e)
                if hasattr(e, 'winerror') and e.winerror == 126:
                    e.strerror = 'Driver module not found.' \
                                 ' Drivers might not be properly installed.'
                log.error(e)

    def change_node_baudrate(self, target_node, vendor_id, product_code,
                             rev_number, serial_number, new_node=None,
                             new_baudrate=None):
        print('\nSwitching slave into CONFIGURATION state...\n')

        bool_result = False
        try:
            bool_result = self.__network.lss.send_switch_state_selective(
                vendor_id,
                product_code,
                rev_number,
                serial_number,
            )
        except Exception as e:
            print('Exception: LSS Timeout. ', e)

        if bool_result:
            if new_baudrate:
                self.__network.lss.configure_bit_timing(
                    CAN_BIT_TIMMING[new_baudrate].value
                )
                sleep(0.1)
            if new_node:
                self.__network.lss.configure_node_id(new_node)
                sleep(0.1)
            self.__network.lss.store_configuration()
            sleep(0.1)
            print('Stored new configuration')
            self.__network.lss.send_switch_state_global(
                self.__network.lss.WAITING_STATE
            )
        else:
            return False

        print('')
        print('Reseting node. Baudrate will be applied after power cycle')
        print('Set properly the baudrate of all the nodes before'
              ' power cycling the devices')
        self.__network.nodes[target_node].nmt.send_command(0x82)

        # Wait until node is reset
        sleep(0.5)

        self.__network.scanner.reset()
        self.__network.scanner.search()
        sleep(0.5)

        for node_id in self.__network.scanner.nodes:
            print('>> Node found: ', node_id)
            node = self.__network.add_node(node_id, self.__eds)

        # Reset all nodes to default state
        self.__network.lss.send_switch_state_global(
            self.__network.lss.WAITING_STATE
        )

        self.__network.nodes[target_node].nmt.start_node_guarding(1)
        return True

    def reset_network(self):
        try:
            self.__network.disconnect()
        except BaseException as e:
            print("Could not reset: Disconnection", e)

        try:
            for node in self.__network.scanner.nodes:
                self.__network.nodes[node].nmt.stop_node_guarding()
            if self.__network.bus:
                self.__network.bus.flush_tx_buffer()
                print("Bus flushed")
        except Exception as e:
            print("Could not stop guarding: ", e)

        try:
            self.__network.connect(bustype=self.__device,
                                   channel=self.__channel,
                                   bitrate=self.__baudrate)
            for node_id in self.__network.scanner.nodes:
                node = self.__network.add_node(node_id, self.__eds)
                node.nmt.start_node_guarding(1)
        except BaseException as e:
            log.warning(e)
            print("Could not reset: Connection", e)

    def detect_nodes(self):
        self.__network.scanner.reset()
        self.__network.scanner.search()
        time.sleep(0.05)
        return self.__network.scanner.nodes

    def scan(self, eds, dict, boot_mode=False, heartbeat=True):
        try:
            self.__network.scanner.reset()
            self.__network.scanner.search()
            time.sleep(0.05)
            for node_id in self.__network.scanner.nodes:
                print("Found node %d!" % node_id)
                node = self.__network.add_node(node_id, eds)

                node.nmt.start_node_guarding(1)

                self.__eds = eds
                self.__dict = dict

                if heartbeat:
                    self.__heartbeat_thread = HearbeatThread(self, node)
                    self.__heartbeat_thread.start()

                self.__servos.append(Servo(self, node, dict,
                                           boot_mode=boot_mode))
        except Exception as e:
            log.error(e)
            print('Exception trying to scan: ', e)

    def connect_through_node(self, eds, dict, node_id, boot_mode=False,
                             heartbeat=True):
        try:
            self.__network.scanner.reset()
            self.__network.scanner.search()
            time.sleep(0.05)

            if node_id in self.__network.scanner.nodes:
                node = self.__network.add_node(node_id, eds)

                node.nmt.start_node_guarding(1)

                self.__eds = eds
                self.__dict = dict

                if heartbeat:
                    self.__heartbeat_thread = HearbeatThread(self, node)
                    self.__heartbeat_thread.start()

                self.__servos.append(Servo(self, node, dict,
                                           boot_mode=boot_mode))
            else:
                log.warning('Node id not found')
        except Exception as e:
            log.error(e)
            print('Exception trying to connect ', e)

    def net_state_subscribe(self, cb):
        """ Subscribe to netowrk state changes.

            Args:
                cb: Callback

            Returns:
                int: Assigned slot.
        """
        r = len(self.__observers)
        self.__observers.append(cb)
        return r

    def stop_heartbeat(self):
        try:
            for node_id, node_obj in self.__network.nodes.items():
                node_obj.nmt.stop_node_guarding()
        except Exception as e:
            print('Could not stop node guarding. ', e)
        if self.__heartbeat_thread is not None and \
                self.__heartbeat_thread.is_alive():
            self.__heartbeat_thread.activate_stop_flag()
            self.__heartbeat_thread.join()
            self.__heartbeat_thread = None

    def disconnect(self):
        try:
            self.stop_heartbeat()
        except Exception as e:
            print('Disconnect: Exception stop_heartbeat(). {}'.format(e))
        try:
            self.__network.disconnect()
        except Exception as e:
            print('Disconnect: Exception network.disconnect(). {}'.format(e))
        try:
            for servo in self.__servos:
                servo.stop_drive_status_thread()
        except Exception as e:
            print('Error stopping DriveStateThread. Exception: {}'.format(e))

    @property
    def servos(self):
        return self.__servos

    @servos.setter
    def servos(self, value):
        self.__servos = value

    @property
    def baudrate(self):
        return self.__baudrate

    @property
    def network(self):
        return self.__network

    @property
    def prot(self):
        """ NET_PROT: Obtain network protocol. """
        return NET_PROT.CAN

    @property
    def net_state(self):
        """ Network state. """
        return self.__net_state

    @net_state.setter
    def net_state(self, new_state):
        self.__net_state = new_state
        for callback in self.__observers:
            callback(self.__net_state)
